#ifndef STIM_BIP_H
#define STIM_BIP_H

#include "../envi/envi_header.h"
#include "../envi/bil.h"
#include "../envi/binary.h"
#include <cstring>
#include <utility>

namespace stim{

/**
	The BIP class represents a 3-dimensional binary file stored using band interleaved by pixel (BIP) image encoding. The binary file is stored
	such that Z-X "frames" are stored sequentially to form an image stack along the y-axis. When accessing the data sequentially on disk,
	the dimensions read, from fastest to slowest, are Z, X, Y.

	This class is optimized for data streaming, and therefore supports extremely large (terabyte-scale) files. Data is loaded from disk
	on request. Functions used to access data are written to support efficient reading.
*/
template <typename T>

class bip: public binary<T> {

protected:
	
	
	std::vector<double> w; //band wavelength
	unsigned int offset;		//header offset

	unsigned long X(){
		return R[1];
	}
	unsigned long Y(){
		return R[2];
	}
	unsigned long Z(){
		return R[0];
	}

	using binary<T>::thread_data;

public:

	using binary<T>::open;
	using binary<T>::file;
	using binary<T>::R;
	using binary<T>::read_line_12;

	/// Open a data file for reading using the class interface.

	/// @param filename is the name of the binary file on disk
	/// @param X is the number of samples along dimension 1
	/// @param Y is the number of samples (lines) along dimension 2
	/// @param B is the number of samples (bands) along dimension 3
	/// @param header_offset is the number of bytes (if any) in the binary header
	/// @param wavelengths is an optional STL vector of size B specifying a numerical label for each band
	bool open(std::string filename, unsigned int X, unsigned int Y, unsigned int B, unsigned int header_offset, std::vector<double> wavelengths){

		//copy the wavelengths to the BSQ file structure
		w = wavelengths;
		//copy the offset to the structure
		offset = header_offset;

		return open(filename, vec<unsigned int>(B, X, Y), header_offset);
		
	}

	/// Retrieve a single band (based on index) and stores it in pre-allocated memory.

	/// @param p is a pointer to an allocated region of memory at least X * Y * sizeof(T) in size.
	/// @param page <= B is the integer number of the band to be copied.
	bool band_index( T * p, unsigned int page){
		return binary<T>::read_plane_0(p, page);
	}

	/// Retrieve a single band (by numerical label) and stores it in pre-allocated memory.

	/// @param p is a pointer to an allocated region of memory at least X * Y * sizeof(T) in size.
	/// @param wavelength is a floating point value (usually a wavelength in spectral data) used as a label for the band to be copied.
	bool band( T * p, double wavelength){

		//if there are no wavelengths in the BSQ file
		if(w.size() == 0)
			return band_index(p, (unsigned int)wavelength);

		unsigned int XY = X() * Y();	//calculate the number of pixels in a band

		unsigned page=0;                      //bands around the wavelength


		//get the bands numbers around the wavelength

		//if wavelength is smaller than the first one in header file
		if ( w[page] > wavelength ){
			band_index(p, page);
			return true;
		}

		while( w[page] < wavelength )
		{
			page++;
			//if wavelength is larger than the last wavelength in header file
			if (page == Z()) {
				band_index(p, Z()-1);
				return true;
			}
		}
		if ( wavelength < w[page] )
		{
			T * p1;
			T * p2;
			p1=(T*)malloc( XY * sizeof(T));                     //memory allocation
			p2=(T*)malloc( XY * sizeof(T));
			band_index(p1, page - 1);
			band_index(p2, page );
			for(unsigned i=0; i < XY; i++){
				double r = (double) (wavelength - w[page-1]) / (double) (w[page] - w[page-1]);
				p[i] = (p2[i] - p1[i]) * r + p1[i];
			}
			free(p1);
			free(p2);
		}
		else                           //if the wavelength is equal to a wavelength in header file
		{
			band_index(p, page);
		}

		return true;
	}

	/// Retrieve a single spectrum (Z-axis line) at a given (x, y) location and stores it in pre-allocated memory.

	/// @param p is a pointer to pre-allocated memory at least B * sizeof(T) in size.
	/// @param x is the x-coordinate (dimension 1) of the spectrum.
	/// @param y is the y-coordinate (dimension 2) of the spectrum.
	bool spectrum(T * p, unsigned x, unsigned y){
		return read_line_12(p, x, y);				//read a line in the binary YZ plane (dimension order for BIP is ZXY)
	}

	/// Retrieves a band of x values from a given xz plane.

	/// @param p is a pointer to pre-allocated memory at least X * sizeof(T) in size
	/// @param c is a pointer to an existing XZ plane (size X*Z*sizeof(T))
	/// @param wavelength is the wavelength of X values to retrieve
	bool read_x_from_xz(T* p, T* c, double wavelength)
	{
		unsigned int B = Z();

		unsigned page=0;                      //samples around the wavelength


		//get the bands numbers around the wavelength

		//if wavelength is smaller than the first one in header file
		if ( w[page] > wavelength ){
			for(unsigned j = 0; j < X(); j++)
			{
				p[j] = c[j * B];
			}		
			return true;
		}

		while( w[page] < wavelength )
		{
			page++;
			//if wavelength is larger than the last wavelength in header file
			if (page == B) {
				for(unsigned j = 0; j < X(); j++)
				{
					p[j] = c[(j + 1) * B - 1];
				}
				return true;
			}
		}
		if ( wavelength < w[page] )
		{
			T * p1;
			T * p2;
			p1=(T*)malloc( X() * sizeof(T));                     //memory allocation
			p2=(T*)malloc( X() * sizeof(T));
			//band_index(p1, page - 1);
			for(unsigned j = 0; j < X(); j++)
			{
				p1[j] = c[j * B + page - 1];
			}
			//band_index(p2, page );
			for(unsigned j = 0; j < X(); j++)
			{
				p2[j] = c[j * B + page];
			}
			
			for(unsigned i=0; i < X(); i++){
				double r = (double) (wavelength - w[page-1]) / (double) (w[page] - w[page-1]);
				p[i] = (p2[i] - p1[i]) * r + p1[i];
			}
			free(p1);
			free(p2);
		}
		else                           //if the wavelength is equal to a wavelength in header file
		{
			//band_index(p, page);
			for(unsigned j = 0; j < X(); j++)
			{
				p[j] = c[j * B + page];
			}
		}

		return true;		
	}

	/// Retrieve a single pixel and stores it in pre-allocated memory.

	/// @param p is a pointer to pre-allocated memory at least sizeof(T) in size.
	/// @param n is an integer index to the pixel using linear array indexing.
	bool pixel(T * p, unsigned n){

		unsigned bandnum = X() * Y();		//calculate numbers in one band
		if ( n >= bandnum){							//make sure the pixel number is right
			std::cout<<"ERROR: sample or line out of range"<<std::endl;
			return false;
		}

		file.seekg(n * Z() * sizeof(T), std::ios::beg);           //point to the certain pixel
		file.read((char *)p, sizeof(T) * Z());
		return true;
	}
	
	//given a Y ,return a ZX slice
	bool read_plane_y(T * p, unsigned y){
		return binary<T>::read_plane_2(p, y);
	}
		
	/// Perform baseline correction given a list of baseline points and stores the result in a new BSQ file.

	/// @param outname is the name of the output file used to store the resulting baseline-corrected data.
	/// @param wls is the list of baseline points based on band labels.	
	bool baseline(std::string outname, std::vector<double> wls){
	
		unsigned N = wls.size();			//get the number of baseline points
		
		std::ofstream target(outname.c_str(), std::ios::binary);	//open the target binary file
		std::string headername = outname + ".hdr";              //the header file name		
		
		//simplify image resolution
		unsigned int ZX = Z() * X();		//calculate the number of points in a Y slice
		unsigned int L = ZX * sizeof(T);			//calculate the number of bytes of a Y slice
		unsigned int B = Z();
		
		T* c;			//pointer to the current Y slice
		c = (T*)malloc(L);  //memory allocation
		
		T* a;			//pointer to the two YZ lines surrounding the current YZ line
		T* b;
		
		a = (T*)malloc(X() * sizeof(T));
		b = (T*)malloc(X() * sizeof(T));


		double ai, bi;	//stores the two baseline points wavelength surrounding the current band
		double ci;		//stores the current band's wavelength
		unsigned control;

		if (a == NULL || b == NULL || c == NULL){
			std::cout<<"ERROR: error allocating memory";
			exit(1);
		}
	//	loop start	correct every y slice
		
		for (unsigned k =0; k < Y(); k++)
		{
			//get the current y slice
			read_plane_y(c, k);
		
			//initialize lownum, highnum, low, high		
			control=0;
			ai = w[0];
			//if no baseline point is specified at band 0,
			//set the baseline point at band 0 to 0
			if(wls[0] != w[0]){
				bi = wls[control];			
				memset(a, (char)0, X() * sizeof(T) );
			}
			//else get the low band
			else{
				control++;
				read_x_from_xz(a, c, ai);
				bi = wls[control];
			}
			//get the high band
			read_x_from_xz(b, c, bi);
		
			//correct every YZ line
			
			for(unsigned cii = 0; cii < B; cii++){
				//update baseline points, if necessary
				if( w[cii] >= bi && cii != B - 1) {
					//if the high band is now on the last BL point?
					if (control != N-1) {
	
						control++;		//increment the index
	
						std::swap(a, b);	//swap the baseline band pointers
	
						ai = bi;
						bi = wls[control];
						read_x_from_xz(b, c, bi);
	
					}
					//if the last BL point on the last band of the file?
					else if ( wls[control] < w[B - 1]) {
	
						std::swap(a, b);	//swap the baseline band pointers
	
						memset(b, (char)0, X() * sizeof(T) );	//clear the high band
	
						ai = bi;
						bi = w[B - 1];
					}
				}

				ci = w[cii];
			
				//perform the baseline correction
				for(unsigned i=0; i < X(); i++)
				{
					double r = (double) (ci - ai) / (double) (bi - ai);
					c[i * B + cii] =(T) ( c[i * B + cii] - (b[i] - a[i]) * r - a[i] );
				}
				
			}//loop for YZ line end  
			target.write(reinterpret_cast<const char*>(c), L);   //write the corrected data into destination	

			thread_data = (double)k / Y() * 100;
		}//loop for Y slice end
		

		
		free(a);
		free(b);
		free(c);
		target.close();

		thread_data = 100;
		return true;	
		
	}
		
	/// Normalize all spectra based on the value of a single band, storing the result in a new BSQ file.

	/// @param outname is the name of the output file used to store the resulting baseline-corrected data.
	///	@param w is the label specifying the band that the hyperspectral image will be normalized to.
	///	@param t is a threshold specified such that a spectrum with a value at w less than t is set to zero. Setting this threshold allows the user to limit division by extremely small numbers.
	bool normalize(std::string outname, double w, double t = 0.0)
	{
		unsigned int B = Z();		//calculate the number of bands
		unsigned int ZX = Z() * X();
		unsigned int XY = X() * Y();	//calculate the number of pixels in a band
		unsigned int S = XY * sizeof(T);		//calculate the number of bytes in a band		
		unsigned int L = ZX * sizeof(T);

		std::ofstream target(outname.c_str(), std::ios::binary);	//open the target binary file
		std::string headername = outname + ".hdr";              //the header file name

		T * c;				//pointer to the current ZX slice
		T * b;				//pointer to the standard band

		b = (T*)malloc( S );     //memory allocation
		c = (T*)malloc( L ); 

		band(b, w);             //get the certain band into memory

		for(unsigned j = 0; j < Y(); j++)
		{
			read_plane_y(c, j);
			unsigned jX = j * X();		//to avoid calculating it many times
			for(unsigned i = 0; i < X(); i++)
			{
				unsigned iB = i * B;		
				for(unsigned m = 0; m < B; m++)
				{
					if( b[i+jX] < t )
						c[m + iB] = (T)0.0;
					else
						c[m + iB] = c[m + iB] / b[i + jX];			//perform normalization
				}								
			}
			target.write(reinterpret_cast<const char*>(c), L);   //write normalized data into destination

			thread_data = (double) j / Y() * 100;
		}

		
		free(b);
		free(c);
		target.close();
		thread_data = 100;
		return true;
	}
	
	/// Convert the current BIP file to a BSQ file with the specified file name.

	/// @param outname is the name of the output BSQ file to be saved to disk.
	bool bsq(std::string outname)
	{	
		std::string temp = outname + "_temp";
		std::string headtemp = temp + ".hdr";
		//first creat a temporary bil file and convert bip file to bil file
		bil(temp);

		stim::bil<T> n;
		if(n.open(temp, X(), Y(), Z(), offset, w)==false){        //open infile
			std::cout<<"ERROR: unable to open input file"<<std::endl;
			exit(1);
		}
		//then convert bil file to bsq file
		n.bsq(outname);
		n.close();
		remove(temp.c_str());
		remove(headtemp.c_str());
		return true;
	}

	/// Convert the current BIP file to a BIL file with the specified file name.

	/// @param outname is the name of the output BIL file to be saved to disk.
	bool bil(std::string outname)
	{
		unsigned int S = X() * Z() * sizeof(T);		//calculate the number of bytes in a ZX slice
		
		std::ofstream target(outname.c_str(), std::ios::binary);
		std::string headername = outname + ".hdr";
		
		T * p;			//pointer to the current ZX slice for bip file
		p = (T*)malloc(S);
		T * q;			//pointer to the current XZ slice for bil file
		q = (T*)malloc(S);
		
		for ( unsigned i = 0; i < Y(); i++)
		{			
			read_plane_y(p, i);
			for ( unsigned k = 0; k < Z(); k++)
			{
				unsigned ks = k * X();
				for ( unsigned j = 0; j < X(); j++)
					q[ks + j] = p[k + j * Z()];

				thread_data = (double)(i * Z() + k) / (Y() * Z()) * 100;
			}
			target.write(reinterpret_cast<const char*>(q), S);   //write a band data into target file	
		}

		thread_data = 100;
		
		free(p);
		free(q);
		target.close();
		return true;
	}

	/// Return a baseline corrected band given two adjacent baseline points and their bands. The result is stored in a pre-allocated array.

	/// @param lb is the label value for the left baseline point
	/// @param rb is the label value for the right baseline point
	/// @param lp is a pointer to an array holding the band image for the left baseline point
	/// @param rp is a pointer to an array holding the band image for the right baseline point
	/// @param wavelength is the label value for the requested baseline-corrected band
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size.
	bool baseline_band(double lb, double rb, T* lp, T* rp, double wavelength, T* result){

		unsigned XY = X() * Y();
		band(result, wavelength);		//get band

		//perform the baseline correction
		double r = (double) (wavelength - lb) / (double) (rb - lb);
		for(unsigned i=0; i < XY; i++){
			result[i] =(T) (result[i] - (rp[i] - lp[i]) * r - lp[i] );
		}
		return true;
	}
	
	/// Return a baseline corrected band given two adjacent baseline points. The result is stored in a pre-allocated array.

	/// @param lb is the label value for the left baseline point
	/// @param rb is the label value for the right baseline point
	/// @param bandwavelength is the label value for the desired baseline-corrected band
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size.
	bool height(double lb, double rb, double bandwavelength, T* result){

		T* lp;
		T* rp;
		unsigned XY = X() * Y();
		unsigned S = XY * sizeof(T);
		lp = (T*) malloc(S);			//memory allocation
		rp = (T*) malloc(S);

		band(lp, lb);
		band(rp, rb);		

		baseline_band(lb, rb, lp, rp, bandwavelength, result);

		free(lp);
		free(rp);
		return true;
	}


	/// Calculate the area under the spectrum between two specified points and stores the result in a pre-allocated array.

	/// @param lb is the label value for the left baseline point
	/// @param rb is the label value for the right baseline point
	/// @param lab is the label value for the left bound (start of the integration)
	/// @param rab is the label value for the right bound (end of the integration)
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size
	bool area(double lb, double rb, double lab, double rab, T* result){

		T* lp;	//left band pointer
		T* rp;	//right band pointer
		T* cur;		//current band 1
		T* cur2;	//current band 2

		unsigned XY = X() * Y();
		unsigned S = XY * sizeof(T);

		lp = (T*) malloc(S);			//memory allocation
		rp = (T*) malloc(S);
		cur = (T*) malloc(S);
		cur2 = (T*) malloc(S);

		memset(result, (char)0, S);

		//find the wavelenght position in the whole band
		unsigned int n = w.size();
		unsigned int ai = 0;		//left bound position
		unsigned int bi = n - 1;		//right bound position



		//to make sure the left and the right bound are in the bandwidth
		if (lb < w[0] || rb < w[0] || lb > w[n-1] || rb >w[n-1]){
			std::cout<<"ERROR: left bound or right bound out of bandwidth"<<std::endl;
			exit(1);
		}
		//to make sure rigth bound is bigger than left bound
		else if(lb > rb){
			std::cout<<"ERROR: right bound should be bigger than left bound"<<std::endl;
			exit(1);
		}

		//get the position of lb and rb
		while (lab >= w[ai]){
			ai++;
		}
		while (rab <= w[bi]){
			bi--;
		}

		band(lp, lb);
		band(rp, rb);

		//calculate the beginning and the ending part
		baseline_band(lb, rb, lp, rp, rab, cur2);		//ending part
		baseline_band(lb, rb, lp, rp, w[bi], cur);
		for(unsigned j = 0; j < XY; j++){
			result[j] += (rab - w[bi]) * (cur[j] + cur2[j]) / 2.0;
		}
		baseline_band(lb, rb, lp, rp, lab, cur2);		//beginnning part
		baseline_band(lb, rb, lp, rp, w[ai], cur);
		for(unsigned j = 0; j < XY; j++){	
			result[j] += (w[ai] - lab) * (cur[j] + cur2[j]) / 2.0;
		}

		//calculate the area
		ai++;
		for(unsigned i = ai; i <= bi ;i++)
		{
			baseline_band(lb, rb, lp, rp, w[ai], cur2);
			for(unsigned j = 0; j < XY; j++)
			{
				result[j] += (w[ai] - w[ai-1]) * (cur[j] + cur2[j]) / 2.0; 
			}
			std::swap(cur,cur2);		//swap the band pointers
		}

		free(lp);
		free(rp);
		free(cur);
		free(cur2);
		return true;
	}

	/// Compute the ratio of two baseline-corrected peaks. The result is stored in a pre-allocated array.

	/// @param lb1 is the label value for the left baseline point for the first peak (numerator)
	/// @param rb1 is the label value for the right baseline point for the first peak (numerator)
	/// @param pos1 is the label value for the first peak (numerator) position
	/// @param lb2 is the label value for the left baseline point for the second peak (denominator)
	/// @param rb2 is the label value for the right baseline point for the second peak (denominator)
	/// @param pos2 is the label value for the second peak (denominator) position
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size
	bool ph_to_ph(double lb1, double rb1, double pos1, double lb2, double rb2, double pos2, T * result){

		T* p1 = (T*)malloc(X() * Y() * sizeof(T));	
		T* p2 = (T*)malloc(X() * Y() * sizeof(T));	
		
		//get the two peak band
		height(lb1, rb1, pos1, p1);
		height(lb2, rb2, pos2, p2);
		//calculate the ratio in result
		for(unsigned i = 0; i < X() * Y(); i++){
			if(p1[i] == 0 && p2[i] ==0)
				result[i] = 1;
			else
				result[i] = p1[i] / p2[i];
		}

		free(p1);
		free(p2);
		return true;	
	}
	
	/// Compute the ratio between a peak area and peak height.

	/// @param lb1 is the label value for the left baseline point for the first peak (numerator)
	/// @param rb1 is the label value for the right baseline point for the first peak (numerator)
	/// @param pos1 is the label value for the first peak (numerator) position
	/// @param lb2 is the label value for the left baseline point for the second peak (denominator)
	/// @param rb2 is the label value for the right baseline point for the second peak (denominator)
	/// @param pos2 is the label value for the second peak (denominator) position
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size
	bool pa_to_ph(double lb1, double rb1, double lab1, double rab1,
					double lb2, double rb2, double pos, T* result){
		
		T* p1 = (T*)malloc(X() * Y() * sizeof(T));	
		T* p2 = (T*)malloc(X() * Y() * sizeof(T));	
		
		//get the area and the peak band
		area(lb1, rb1, lab1, rab1, p1);
		height(lb2, rb2, pos, p2);
		//calculate the ratio in result
		for(unsigned i = 0; i < X() * Y(); i++){
			if(p1[i] == 0 && p2[i] ==0)
				result[i] = 1;
			else
				result[i] = p1[i] / p2[i];
		}

		free(p1);
		free(p2);
		return true;	
	}		
	
	/// Compute the ratio between two peak areas.

	/// @param lb1 is the label value for the left baseline point for the first peak (numerator)
	/// @param rb1 is the label value for the right baseline point for the first peak (numerator)
	/// @param lab1 is the label value for the left bound (start of the integration) of the first peak (numerator)
	/// @param rab1 is the label value for the right bound (end of the integration) of the first peak (numerator)
	/// @param lb2 is the label value for the left baseline point for the second peak (denominator)
	/// @param rb2 is the label value for the right baseline point for the second peak (denominator)
	/// @param lab2 is the label value for the left bound (start of the integration) of the second peak (denominator)
	/// @param rab2 is the label value for the right bound (end of the integration) of the second peak (denominator)	
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size
	bool pa_to_pa(double lb1, double rb1, double lab1, double rab1,
					double lb2, double rb2, double lab2, double rab2, T* result){
		
		T* p1 = (T*)malloc(X() * Y() * sizeof(T));	
		T* p2 = (T*)malloc(X() * Y() * sizeof(T));	
		
		//get the area and the peak band
		area(lb1, rb1, lab1, rab1, p1);
		area(lb2, rb2, lab2, rab2, p2);
		//calculate the ratio in result
		for(unsigned i = 0; i < X() * Y(); i++){
			if(p1[i] == 0 && p2[i] ==0)
				result[i] = 1;
			else
				result[i] = p1[i] / p2[i];
		}

		free(p1);
		free(p2);
		return true;	
	}		

	/// Compute the definite integral of a baseline corrected peak.

	/// @param lb is the label value for the left baseline point
	/// @param rb is the label value for the right baseline point
	/// @param lab is the label for the start of the definite integral
	/// @param rab is the label for the end of the definite integral
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size
	bool x_area(double lb, double rb, double lab, double rab, T* result){
		T* lp;	//left band pointer
		T* rp;	//right band pointer
		T* cur;		//current band 1
		T* cur2;	//current band 2

		unsigned XY = X() * Y();
		unsigned S = XY * sizeof(T);

		lp = (T*) malloc(S);			//memory allocation
		rp = (T*) malloc(S);
		cur = (T*) malloc(S);
		cur2 = (T*) malloc(S);

		memset(result, (char)0, S);

		//find the wavelenght position in the whole band
		unsigned int n = w.size();
		unsigned int ai = 0;		//left bound position
		unsigned int bi = n - 1;		//right bound position

		//to make sure the left and the right bound are in the bandwidth
		if (lb < w[0] || rb < w[0] || lb > w[n-1] || rb >w[n-1]){
			std::cout<<"ERROR: left bound or right bound out of bandwidth"<<std::endl;
			exit(1);
		}
		//to make sure rigth bound is bigger than left bound
		else if(lb > rb){
			std::cout<<"ERROR: right bound should be bigger than left bound"<<std::endl;
			exit(1);
		}

		//get the position of lb and rb
		while (lab >= w[ai]){
			ai++;
		}
		while (rab <= w[bi]){
			bi--;
		}

		band(lp, lb);
		band(rp, rb);

		//calculate the beginning and the ending part
		baseline_band(lb, rb, lp, rp, rab, cur2);		//ending part
		baseline_band(lb, rb, lp, rp, w[bi], cur);
		for(unsigned j = 0; j < XY; j++){
			result[j] += (rab - w[bi]) * (rab + w[bi]) * (cur[j] + cur2[j]) / 4.0;
		}
		baseline_band(lb, rb, lp, rp, lab, cur2);		//beginnning part
		baseline_band(lb, rb, lp, rp, w[ai], cur);
		for(unsigned j = 0; j < XY; j++){	
			result[j] += (w[ai] - lab) * (w[ai] + lab) * (cur[j] + cur2[j]) / 4.0;
		}

		//calculate f(x) times x
		ai++;
		for(unsigned i = ai; i <= bi ;i++)
		{
			baseline_band(lb, rb, lp, rp, w[ai], cur2);
			for(unsigned j = 0; j < XY; j++)
			{
				result[j] += (w[ai] - w[ai-1]) * (w[ai] + w[ai-1]) * (cur[j] + cur2[j]) / 4.0; 
			}
			std::swap(cur,cur2);		//swap the band pointers
		}

		free(lp);
		free(rp);
		free(cur);
		free(cur2);
		return true;
	}

	/// Compute the centroid of a baseline corrected peak.

	/// @param lb is the label value for the left baseline point
	/// @param rb is the label value for the right baseline point
	/// @param lab is the label for the start of the peak
	/// @param rab is the label for the end of the peak
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size
	bool cpoint(double lb, double rb, double lab, double rab, T* result){
		T* p1 = (T*)malloc(X() * Y() * sizeof(T));	
		T* p2 = (T*)malloc(X() * Y() * sizeof(T));	
		
		//get the area and the peak band
		x_area(lb, rb, lab, rab, p1);
		area(lb, rb, lab, rab, p2);
		//calculate the ratio in result
		for(unsigned i = 0; i < X() * Y(); i++){
			if(p1[i] == 0 && p2[i] ==0)
				result[i] = 1;
			else
				result[i] = p1[i] / p2[i];
		}

		free(p1);
		free(p2);
		return true;	
	}

	/// Create a mask based on a given band and threshold value.

	/// All pixels in the
	/// specified band greater than the threshold are true and all pixels less than the threshold are false.
	/// @param mask_band is the band used to specify the mask
	/// @param threshold is the threshold used to determine if the mask value is true or false
	/// @param p is a pointer to a pre-allocated array at least X * Y in size
	bool build_mask(double mask_band, double threshold, unsigned char* p){

		T* temp = (T*)malloc(X() * Y() * sizeof(T));		//allocate memory for the certain band
		band(temp, mask_band);

		for (unsigned int i = 0; i < X() * Y();i++) {
				if (temp[i] < threshold)
					p[i] = 0;
				else
					p[i] = 255;
		}

		free(temp);
		return true;

	}

	/// Apply a mask file to the BSQ image, setting all values outside the mask to zero.

	/// @param outfile is the name of the masked output file
	/// @param p is a pointer to memory of size X * Y, where p(i) = 0 for pixels that will be set to zero.
	bool apply_mask(std::string outfile, unsigned char* p){

		std::ofstream target(outfile.c_str(), std::ios::binary);

		unsigned ZX = Z() * X();		//calculate the number of values in a page (XZ in BIP)
		unsigned L = ZX * sizeof(T);	//calculate the number of bytes in a page

		T * temp = (T*)malloc(L);		//allocate space for that page

		for (unsigned i = 0; i < Y(); i++)			//for each page (Y in BIP)
		{
			read_plane_y(temp, i);							//load that page (it's pointed to by temp)
			for ( unsigned j = 0; j < X(); j++)	//for each X value
			{
				for (unsigned k = 0; k < Z(); k++)	//for each B value (band)
				{
					if (p[i * X() + j] == 0)	//if the mask value is zero
					temp[j * Z() + k] = 0;			//set the pixel value to zero
				else								//otherwise just continue
					continue;
				}
			}
			target.write(reinterpret_cast<const char*>(temp), L);   //write the edited band data into target file
		}
		target.close();						//close the target file
		free(temp);							//free allocated memory
		return true;						//return true
	}


	/// Saves to disk only those spectra corresponding to mask values != 0
	bool sift(std::string outfile, unsigned char* mask){

		//reset the file pointer to the beginning of the file
		file.seekg(0, std::ios::beg);		

		// open an output stream
		std::ofstream target(outfile.c_str(), std::ios::binary);

		//allocate space for a single spectrum
		unsigned long B = Z();
		T* spectrum = (T*) malloc(B * sizeof(T));

		//calculate the number of pixels in a band
		unsigned long XY = X() * Y();

		//for each pixel
		unsigned long skip = 0;					//number of spectra to skip
		for(unsigned long x = 0; x < XY; x++){

			//if the current pixel isn't masked
			if( mask[x] == 0){
				//increment the number of skipped pixels
				skip++;
			}
			//if the current pixel is masked
			else{

				//skip the intermediate pixels
				file.seekg(skip * B * sizeof(T), std::ios::cur);

				//set the skip value to zero
				skip = 0;

				//read this pixel into memory
				file.read((char *)spectrum, B * sizeof(T));

				//write this pixel out
				target.write((char *)spectrum, B * sizeof(T));

				thread_data = (double) x / XY * 100;
			}

		}

		//close the output file
		target.close();
		free(spectrum);

		thread_data = 100;

		return true;
	}

	bool unsift(std::string outfile, unsigned char* mask, unsigned int samples, unsigned int lines){

		// open an output stream
		std::ofstream target(outfile.c_str(), std::ios::binary);

		//reset the file pointer to the beginning of the file
		file.seekg(0, std::ios::beg);		

		//allocate space for a single spectrum
		unsigned long B = Z();
		T* spectrum = (T*) malloc(B * sizeof(T));

		//allocate space for a spectrum of zeros
		T* zeros = (T*) malloc(B * sizeof(T));
		memset(zeros, 0, B * sizeof(T));

		//calculate the number of pixels in a band
		unsigned long XY = samples * lines;

		//for each pixel
		unsigned long skip = 0;					//number of spectra to skip
		for(unsigned long x = 0; x < XY; x++){

			//if the current pixel isn't masked
			if( mask[x] == 0){

				//write a bunch of zeros
				target.write((char *)zeros, B * sizeof(T));
			}
			//if the current pixel is masked
			else{

				//read a pixel into memory
				file.read((char *)spectrum, B * sizeof(T));

				//write this pixel out
				target.write((char *)spectrum, B * sizeof(T));
			}

			thread_data = (double)x / XY * 100;

		}

		//close the output file
		target.close();
		free(spectrum);

		thread_data = 100;

		return true;


	}



	/// @param p is a pointer to memory of size X * Y * sizeof(T) that will store the band averages.
	bool band_avg(T* p){
		unsigned long long XY = X() * Y();
		//get every pixel and calculate average value
		T* temp = (T*)malloc(sizeof(T) * Z());
		T sum;
		for (unsigned i = 0; i < XY; i++){
			pixel(temp, i);
			//calculate the sum value of every value
			sum = 0;		//initialize sum value
			for (unsigned j = 0; j < Z(); j++){
				sum += temp[j]/(T)Z();
			}
			p[i] = sum;
		}
		free(temp);
		return true;
	}

	/// Calculate the mean value for all masked (or valid) pixels in a band and returns the average spectrum

	/// @param p is a pointer to pre-allocated memory of size [B * sizeof(T)] that stores the mean spectrum
	/// @param mask is a pointer to memory of size [X * Y] that stores the mask value at each pixel location
	bool avg_band(T*p, unsigned char* mask){
		unsigned long long XY = X() * Y();
		T* temp = (T*)malloc(sizeof(T) * Z());
		//Iinitialize
		for (unsigned j = 0; j < Z(); j++){
			p[j] = 0;
		}
		//calculate vaild number in a band
		unsigned count = 0;
		for (unsigned j = 0; j < XY; j++){
			if (mask[j] != 0){
				count++;
			}
		}
		//calculate average number of a band
		for (unsigned i = 0; i < XY; i++){
			if (mask[i] != 0){			
				pixel(temp, i);
				for (unsigned j = 0; j < Z(); j++){
					p[j] += temp[j] / (T)count;
				}
			}
		}
		free(temp);
		return true;
	}
	
	/// Calculate the covariance matrix for all masked pixels in the image.

	/// @param co is a pointer to pre-allocated memory of size [B * B] that stores the resulting covariance matrix
	/// @param avg is a pointer to memory of size B that stores the average spectrum
	/// @param mask is a pointer to memory of size [X * Y] that stores the mask value at each pixel location
	bool co_matrix(T* co, T* avg, unsigned char *mask){
		//memory allocation
		unsigned long long xy = X() * Y();
		unsigned int B = Z();
		T* temp = (T*)malloc(sizeof(T) * B);
		//count vaild pixels in a band
		unsigned count = 0;
		for (unsigned j = 0; j < xy; j++){
			if (mask[j] != 0){
				count++;
			}
		}
		//initialize correlation matrix
		for (unsigned i = 0; i < B; i++){
			for (unsigned k = 0; k < B; k++){
				co[i * B + k] = 0;
			}
		}
		//calculate correlation coefficient matrix
		for (unsigned j = 0; j < xy; j++){
			if (mask[j] != 0){
				pixel(temp, j);
				for (unsigned i = 0; i < B; i++){
					for (unsigned k = i; k < B; k++){
						co[i * B + k] += (temp[i] - avg[i]) * (temp[k] - avg[k]) / count;
					}
				}
			}
		}
		//because correlation matrix is symmetric
		for (unsigned i = 0; i < B; i++){
			for (unsigned k = i + 1; k < B; k++){
				co[k * B + i] = co[i * B + k];
			}
		}

		free(temp);
		return true;
	}


	/// Crop a region of the image and save it to a new file.

	/// @param outfile is the file name for the new cropped image
	/// @param x0 is the lower-left x pixel coordinate to be included in the cropped image
	/// @param y0 is the lower-left y pixel coordinate to be included in the cropped image
	/// @param x1 is the upper-right x pixel coordinate to be included in the cropped image
	/// @param y1 is the upper-right y pixel coordinate to be included in the cropped image
	bool crop(std::string outfile, unsigned long long x0,
								   unsigned long long y0,
								   unsigned long long x1,
								   unsigned long long y1,
								   unsigned long long b0,
								   unsigned long long b1){

		//calculate the new number of samples, lines, and bands
		unsigned long long samples = x1 - x0;
		unsigned long long lines = y1 - y0;
		unsigned long long bands = b1 - b0;

		//calculate the length of one cropped spectrum
		unsigned long long L = bands * sizeof(T);

		//unsigned long long L = Z() * sizeof(T);

		//allocate space for the spectrum
		T* temp = (T*)malloc(L);

		//open an output file for binary writing
		std::ofstream out(outfile.c_str(), std::ios::binary);

		//seek to the first pixel in the cropped image
		file.seekg( (y0 * X() * Z() + x0 * Z() + b0) * sizeof(T), std::ios::beg);

		//distance between sample spectra in the same line
		unsigned long long jump_sample = ( (Z() - b1) + b0 ) * sizeof(T);

		//distance between sample spectra in adjacent lines
		unsigned long long jump_line = (X() - x1) * Z() * sizeof(T);


		//unsigned long long sp = y0 * X() + x0;		//start pixel

		//for each pixel in the image
		for (unsigned y = 0; y < lines; y++)
		{
			for (unsigned x = 0; x < samples; x++)
			{
				//read the cropped spectral region
				file.read( (char*) temp, L );
				//pixel(temp, sp + x + y * X());
				out.write(reinterpret_cast<const char*>(temp), L);   //write slice data into target file	

				file.seekg(jump_sample, std::ios::cur);

				thread_data = (double)(y * samples + x) / (lines * samples) * 100;
			}

			file.seekg(jump_line, std::ios::cur);
		}
		free(temp);

		thread_data = 100;
		return true;
	}


	/// Close the file.
	bool close(){
		file.close();
		return true;
	}

	};
}

#endif
